/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/

#include <string.h>
#include <stdint.h>
#include "printf.h"
#include "entry.h"
#include "util.h"
#include "load.h"
#include "vfs.h"

#define VERBOSE							0

//ultrix.45 stock is 256. 64 works well. 32 hangs due to various calculations over & underflowing
#define ADJUST_HZ_REAL_TO						16
#define ADJUST_HZ_REPORTED_TO					64


#if defined(ADJUST_HZ_REAL_TO) && ((ADJUST_HZ_REAL_TO & (ADJUST_HZ_REAL_TO - 1)) || ADJUST_HZ_REAL_TO < 2 || ADJUST_HZ_REAL_TO > 8192)
	#error "ADJUST_HZ_REAL_TO must be a power of 2 between 2 and 8192"
#endif

#if defined(ADJUST_HZ_REPORTED_TO) && ((ADJUST_HZ_REPORTED_TO & (ADJUST_HZ_REPORTED_TO - 1)) || ADJUST_HZ_REPORTED_TO < 2 || ADJUST_HZ_REPORTED_TO > 8192)
	#error "ADJUST_HZ_REPORTED_TO must be a power of 2 between 2 and 8192"
#endif

struct CoffHdr {
	uint16_t magic;			//0x00
	uint16_t nSects;		//0x02
	uint32_t timestamp;		//0x04
	uint32_t symtabOffst;	//0x08
	uint32_t nSymbols;		//0x0c
	uint16_t optHdrSz;		//0x10
	uint16_t flags;			//0x12
};

#define MIPS_COFF_LE_MAGIC				0x162

#define COFF_FILE_FLAG_RELOCATABLE		0x0001
#define COFF_FILE_FLAG_EXECUTABLE		0x0002
#define COFF_FILE_FLAG_NO_LINE_NUMS		0x0004
#define COFF_FILE_FLAG_NO_SYMBOLS		0x0008


struct UltrixOptionalHdr {	//not very optional
	uint16_t magic;			//0x00
	uint16_t version;		//0x02
	uint32_t textSz;		//0x04
	uint32_t dataSz;		//0x08
	uint32_t bssSz;			//0x0c
	uint32_t entrypt;		//0x10
	uint32_t textBase;		//0x14
	uint32_t dataBase;		//0x18
	uint32_t bssBase;		//0x1c
	int32_t gprMask;		//0x20
	int32_t fprMask[4];		//0x24
	uint32_t gpValue;		//0x34
};	//0x38

#define ULTRIX_OPT_HDR_MAGIC			0x0107
#define ULTRIX_OPT_HDR_MIN_VER			0x0300

struct CoffSection {
	char name[8];
	uint32_t pa, va, size,	fileDataOfst, fileRelocOfst, fileLineNoOfst;
	uint16_t nRelocs, nLineNoEntries;
	uint32_t flags;
};


#define COFF_SECT_FLAG_TEXT				0x00000020
#define COFF_SECT_FLAG_DATA				0x00000040
#define COFF_SECT_FLAG_BSS				0x00000080

struct CoffMipsSymTabTab {
	uint16_t magic;					//0x00
	uint16_t version;				//0x02
	
	//line numbers
	uint32_t lineNoEntriesCount;	//0x04
	uint32_t lineNoEntriesBytes;	//0x08
	uint32_t lineNoEntriesOffset;	//0x0c
	
	//dense numbers table
	uint32_t denseNumsCount;		//0x10
	uint32_t denseNumsOffset;		//0x14
	
	//procedures
	uint32_t proceduresCount;		//0x18
	uint32_t proceduresOffset;		//0x1c
	
	//local symbols
	uint32_t localSymsCount;		//0x20
	uint32_t localSymsOffset;		//0x24
	
	//opt symbols
	uint32_t optSymsCount;			//0x28
	uint32_t optSymsOffset;			//0x2c
	
	//aux symbols
	uint32_t auxSymsCount;			//0x30
	uint32_t auxSymsOffset;			//0x34
	
	//local strings
	uint32_t localStringsBytes;		//0x38
	uint32_t localStringOffset;		//0x3c
	
	//external strings
	uint32_t externalStringsBytes;	//0x40
	uint32_t externalStringOffset;	//0x44
	
	//file descriptors
	uint32_t fileDescriptorsCount;	//0x48
	uint32_t fileDescriptorsOffset;	//0x4c
	
	//relative file descriptors
	uint32_t relFileDescrsCount;	//0x50
	uint32_t relFileDescrsOffset;	//0x54
	
	//external symbols
	uint32_t extSymbolsCount;		//0x58
	uint32_t extSymbolsOffset;		//0x5c
};

#define COFF_MIPS_SYMTABTAB_MAGIC		0x7009
#define COFF_MIPS_SYMTABTAB_MIN_VER		0x0300

struct CoffMipsFileDescr {
	uint32_t startAddr;				//0x00
	uint32_t fileNameOffset;		//0x04
	uint32_t fileStringsOffset;		//0x08
	uint32_t fileStringsBytes;		//0x0c
	uint32_t fileSymsBase;			//0x10
	uint32_t fileSymsCount;			//0x14
	uint32_t fileLineSymsBase;		//0x18
	uint32_t fileLineSymsCount;		//0x1c
	uint32_t fileOptOptsBase;		//0x20
	uint32_t fileOptOptsCount;		//0x24
	uint16_t firstProcedureIdx;		//0x28
	uint16_t numProcedures;			//0x2a
	uint32_t fileAuxBase;			//0x2c
	uint32_t fileAuxCount;			//0x30
	uint32_t fileIndirectTblBase;	//0x34
	uint32_t fileIndirectTblCount;	//0x38
	uint32_t flags;					//0x3c
	uint32_t fileLineNosBase;		//0x40
	uint32_t fileLineNosCount;		//0x44
};

struct CoffMipsProcedureDescr {
	uint32_t addr;					//0x00
	uint32_t symEntriesStart;		//0x04
	uint32_t lineNosStart;			//0x08
	uint32_t regMask;				//0x0c
	uint32_t regOffst;				//0x10
	uint32_t optOptsStart;			//0x14
	uint32_t fregMask;				//0x18
	uint32_t fregOffst;				//0x1c
	uint32_t frameOffst;			//0x20
	uint16_t frameReg;				//0x24
	uint16_t pcReg;					//0x26
	uint32_t minLine;				//0x28
	uint32_t maxLine;				//0x2c
	uint32_t lineNosOffst;			//0x30
};

struct CoffMipsSymbolEntry {
	uint32_t stringOfst;			//0x00
	uint32_t addr;					//0x04
	uint32_t symbolTyp			: 6;//0x08
	uint32_t symStorageClass	: 5;
	uint32_t rfu				: 1;
	uint32_t symbolIdx			: 20;
};

#define COFF_MIPS_SYMTYPE_NONE		0
#define COFF_MIPS_SYMTYPE_GLOBAL	1
#define COFF_MIPS_SYMTYPE_STATIC	2
#define COFF_MIPS_SYMTYPE_PARAM		3
#define COFF_MIPS_SYMTYPE_LOCAL		4
#define COFF_MIPS_SYMTYPE_LABEL		5
#define COFF_MIPS_SYMTYPE_PROCEDURE	6
#define COFF_MIPS_SYMTYPE_BLOCK		7
#define COFF_MIPS_SYMTYPE_END		8
#define COFF_MIPS_SYMTYPE_MEMBER	9
#define COFF_MIPS_SYMTYPE_TYPEDEF	10
#define COFF_MIPS_SYMTYPE_FILE		11




/* this was a valiant efford, and to worked on the installer kernel that had symbols, but release kernel does not so we no longer use this

	//do this before we do any loading as this will use up a lot of ram starting at the very top of our RAM
	//this is much faster than seeking through the file
	//but worry not, this is still rather slow to do...
	static bool ultrixLocatePatchSitesBySymbols(struct VfsFile *fil, const struct CoffHdr *chdr, const struct UltrixOptionalHdr *oh, uint32_t fileSize, const char **targetNames, uint32_t *targetAddrs, uint32_t nTargets)
	{
		if (1)
			return false;
		
		struct CoffMipsFileDescr *file, *files = (struct CoffMipsFileDescr*)stack_start;
		struct CoffMipsProcedureDescr *procs;
		struct CoffMipsSymbolEntry *syms;
		struct CoffMipsSymTabTab stt;
		uint32_t i, targetsLeft;
		char *strings;
		
		if (nTargets > 32) {
			
			pr("cannot handle this many patchsites\n");
			return false;
		}
		targetsLeft = (1 << nTargets) - 1;
		
		pr("PATCHING BUGS...\n");
		pr(" reading symbol table summary\n");
		if (!vfsFileSeek(fil, chdr->symtabOffst) || sizeof(stt) != vfsFileRead(fil, &stt, sizeof(stt))) {
			
			pr("cannot read symbol table table\n");
			return false;
		}
		
		if (VERBOSE) {
			pr("STT:\r\n");
			pr(" lineNoEntriesCount 0x%08x\r\n", stt.lineNoEntriesCount);
			pr(" lineNoEntriesBytes 0x%08x\r\n", stt.lineNoEntriesBytes);
			pr(" lineNoEntriesOffset 0x%08x\r\n", stt.lineNoEntriesOffset);
			
			pr(" denseNumsCount 0x%08x\r\n", stt.denseNumsCount);
			pr(" denseNumsOffset 0x%08x\r\n", stt.denseNumsOffset);
			
			pr(" proceduresCount 0x%08x\r\n", stt.proceduresCount);
			pr(" proceduresOffset 0x%08x\r\n", stt.proceduresOffset);
			
			pr(" localSymsCount 0x%08x\r\n", stt.localSymsCount);
			pr(" localSymsOffset 0x%08x\r\n", stt.localSymsOffset);
			
			pr(" optSymsCount 0x%08x\r\n", stt.optSymsCount);
			pr(" optSymsOffset 0x%08x\r\n", stt.optSymsOffset);
			
			pr(" auxSymsCount 0x%08x\r\n", stt.auxSymsCount);
			pr(" auxSymsOffset 0x%08x\r\n", stt.auxSymsOffset);
			
			pr(" localStringsBytes 0x%08x\r\n", stt.localStringsBytes);
			pr(" localStringOffset 0x%08x\r\n", stt.localStringOffset);
			
			pr(" externalStringsBytes 0x%08x\r\n", stt.externalStringsBytes);
			pr(" externalStringOffset 0x%08x\r\n", stt.externalStringOffset);
			
			pr(" fileDescriptorsCount 0x%08x\r\n", stt.fileDescriptorsCount);
			pr(" fileDescriptorsOffset 0x%08x\r\n", stt.fileDescriptorsOffset);
			
			pr(" relFileDescrsCount 0x%08x\r\n", stt.relFileDescrsCount);
			pr(" relFileDescrsOffset 0x%08x\r\n", stt.relFileDescrsOffset);
			
			pr(" extSymbolsCount 0x%08x\r\n", stt.extSymbolsCount);
			pr(" extSymbolsOffset 0x%08x\r\n", stt.extSymbolsOffset);
		}
	
		
		procs = (struct CoffMipsProcedureDescr*)(files + stt.fileDescriptorsCount);
		syms = (struct CoffMipsSymbolEntry*)(procs + stt.proceduresCount);
		strings = (char*)(syms + stt.localSymsCount);
		
		if (!stt.proceduresCount || !stt.localSymsCount || !stt.localStringsBytes || !stt.fileDescriptorsCount) {
			
			pr("This kernel lacks symbols and cannot be analyzed symbolically\n");
			return false;
		}
		
		pr("Reading %u bytes of the %s table\n", stt.proceduresCount * sizeof(*procs), "procedure");
		if (!vfsFileSeek(fil, stt.proceduresOffset) || !copyWithProgress(procs, fil, stt.proceduresCount * sizeof(*procs))) {
			
			pr("cannot read in procedures table\n");
			return false;
		}
		
		pr("Reading %u bytes of the %s table\n", stt.localSymsCount * sizeof(*syms), "symbol");
		if (!vfsFileSeek(fil, stt.localSymsOffset) || !copyWithProgress(syms, fil, stt.localSymsCount * sizeof(*syms))) {
			
			pr("cannot read in local symbols table\n");
			return false;
		}
	
		pr("Reading %u bytes of the %s table\n", stt.localStringsBytes, "string");
		if (!vfsFileSeek(fil, stt.localStringOffset) || !copyWithProgress(strings, fil, stt.localStringsBytes)) {
			
			pr("cannot read in local strings table\n");
			return false;
		}
		
		pr("Reading %u bytes of the %s table\n", stt.fileDescriptorsCount * sizeof(*files), "file");
		if (!vfsFileSeek(fil, stt.fileDescriptorsOffset) || !copyWithProgress(files, fil, stt.fileDescriptorsCount * sizeof(*files))) {
			
			pr("cannot read in file descriptors table\n");
			return false;
		}
		
		pr(" analyzing...\n");
		for (file = files, i = 0; i < stt.fileDescriptorsCount; i++, file++) {
			
			struct CoffMipsProcedureDescr *proc = procs + file->firstProcedureIdx;
			char *fileStrings = strings + file->fileStringsOffset;
			uint32_t j;
			
			for (j = 0; j < file->numProcedures; j++, proc++) {
				
				uint32_t numSyms = (j == file->numProcedures - 1) ? file->fileSymsCount - proc->symEntriesStart : proc[1].symEntriesStart - proc->symEntriesStart;
				struct CoffMipsSymbolEntry *sym = syms + proc->symEntriesStart + file->fileSymsBase;
				uint32_t k, m;
				
	//			pr("  proc %u (abs %u):\n", j, j + file->firstProcedureIdx);
	//			pr("      func @ 0x%08x\n", proc->addr + file->startAddr);
	//			pr("      symbols %u + %u\n", proc->symEntriesStart, numSyms);
				
				for (k = 0; k < numSyms; k++, sym++) {
					
					uint32_t va = sym->addr + ((sym->symbolTyp == COFF_MIPS_SYMTYPE_PROCEDURE) ? 0 : proc->addr + file->startAddr);
					const char *thisFuncName;
					
	//				pr("         sym %u / %u (absabs %u)\n", k, numSyms, sym - syms);
	//				if (sym->stringOfst)
	//					pr("         name '%s'\n", fileStrings + sym->stringOfst);
	//				pr("         addr 0x%08x\n", va);
	//				pr("         type %d\n", sym->symbolTyp );
					
					//could it be what we seek (the symbol that signifies this procedure itself)?
					//there will only be one "procedure" type symbol per procedure, so if we see it, do not go on
					if (sym->symbolTyp != COFF_MIPS_SYMTYPE_PROCEDURE || !sym->stringOfst)
						continue;
					
					thisFuncName = fileStrings + sym->stringOfst;
					
					for (m = 0; m < nTargets; m++) {
						
						if (!(targetsLeft & (1 << m)))
							continue;
						
						if (strcmp(thisFuncName, targetNames[m]))
							continue;
						
						pr("  found '%s' @ 0x%08x\n", targetNames[m], va);
						targetAddrs[m] = va;
						targetsLeft &=~ (1 << m);
						if (!targetsLeft)
							return true;
						break;
					}
				}	
			}
			
		}
		
		pr(" failed...\n");
		return false;
	}
	
	it used to be called like this:
	
		static const char *patchNames[] = {"r3_kn01flush_cache", "microdelay"};
		uint32_t patchSites[sizeof(patchNames) / sizeof(*patchNames)];
		
		if (!ultrixLocatePatchSitesBySymbols(f, &chdr, &oh, fileSize, patchNames, patchSites, sizeof(patchNames) / sizeof(*patchNames))) {
			panic("bad");
		
		
		//patch "r3_kn01flush_cache"
		((uint32_t*)patchSites[0])[0] = 0x03e00008;	//JR $ra
		((uint32_t*)patchSites[0])[1] = 0x00000000;	//NOP
		
		//patch "microdelay" which would call "kn01delay" - our virtual hardware needs no delays
		((uint32_t*)patchSites[1])[0] = 0x03e00008;	//JR $ra
		((uint32_t*)patchSites[1])[1] = 0x00000000;	//NOP
	
	
*/

#define EXPECT_ZERO			0
#define EXPECT_NOP			(EXPECT_ZERO)
#define EXPECT_JAL			1
#define EXPECT_GP_LOAD		2
#define EXPECT_WORD			3
#define EXPECT_ANYTHING		4	//literally any word will do
#define EXPECT_DONE			5
#define EXPECT_WORD_VAL(x)	EXPECT_WORD, (((x) >> 0) & 0xff), (((x) >> 8) & 0xff), (((x) >> 16) & 0xff), (((x) >> 24) & 0xff)

struct unalignedU32 {
	uint32_t v;
} __attribute__((packed));

static uint32_t* ultrixPatchFind(uint32_t *code, uint32_t nWordsMinusPatLen, const uint8_t *pattern)
{
	uint32_t *ret = NULL, *codeP;
	const uint8_t *pp;
	uint_fast8_t cp;
	uint32_t i;
	
	for (i = 0; i < nWordsMinusPatLen; i++, code++) {
		
		codeP = code;
		pp = pattern;
		while (1) switch (cp = *pp++) {
			
			case EXPECT_ZERO:
				if (*codeP++ != 0x00000000)
					goto nomatch;
				break;
			
			case EXPECT_JAL:
				if ((*codeP++ >> 26) != 3)
					goto nomatch;
				break;
			
			case EXPECT_GP_LOAD:
				if ((*codeP++ >> 21) != 0b10001111100)
					goto nomatch;
				break;
			
			case EXPECT_WORD:
				if (*codeP++ != ((struct unalignedU32*)pp)->v)
					goto nomatch;
				pp += 4;
				break;
			
			case EXPECT_ANYTHING:
				codeP++;
				break;
			
			case EXPECT_DONE:
				goto end;
			
			default:
				__builtin_unreachable();
		}
end:
		//reached only if we match
		if (!ret)
			ret = code;
		else	//dupe
			return NULL;
	
nomatch:;
	}
	
	return ret;
}


static bool ultrixPatchCode(uint32_t *code, uint32_t nWords)
{
	static const uint8_t flush_cache[] = {EXPECT_WORD_VAL(0x27bdffd8), EXPECT_WORD_VAL(0xafbf001c), EXPECT_JAL, EXPECT_NOP, EXPECT_GP_LOAD, EXPECT_WORD_VAL(0xafa20024), EXPECT_WORD_VAL(0x8dcf006c), EXPECT_NOP, EXPECT_WORD_VAL(0x01e0f809), EXPECT_NOP, EXPECT_GP_LOAD, EXPECT_NOP, EXPECT_WORD_VAL(0x8f190028), EXPECT_NOP, EXPECT_WORD_VAL(0x0320f809), EXPECT_NOP, EXPECT_WORD_VAL(0x8fa40024), EXPECT_JAL, EXPECT_NOP, EXPECT_WORD_VAL(0x8fbf001c), EXPECT_WORD_VAL(0x27bd0028),  EXPECT_WORD_VAL(0x03e00008), EXPECT_NOP, EXPECT_DONE};
	static const uint8_t microdelay[] = {EXPECT_WORD_VAL(0x27bdffe0), EXPECT_WORD_VAL(0xafbf001c), EXPECT_GP_LOAD, EXPECT_NOP, EXPECT_WORD_VAL(0x8dcf0038), EXPECT_NOP, EXPECT_WORD_VAL(0x01e0f809), EXPECT_NOP, EXPECT_WORD_VAL(0x8fbf001c), EXPECT_WORD_VAL(0x27bd0020), EXPECT_WORD_VAL(0x03e00008), EXPECT_NOP, EXPECT_DONE};
	uint32_t *at;
	
	at = ultrixPatchFind(code, nWords - 23, flush_cache);
	if (at) {
		pr("flush_cache @ 0x%08x\n", at);
		at[0] = 0x03e00008;	//JR $ra
		at[1] = 0x00000000;	//NOP
	}
	else
		return false;

	at = ultrixPatchFind(code, nWords - 12, microdelay);
	if (at) {
		pr("microdelay @ 0x%08x\n", at);
		at[0] = 0x03e00008;	//JR $ra
		at[1] = 0x00000000;	//NOP
	}
	else
		return false;
	
	#ifdef ADJUST_HZ_REAL_TO
	
		static const uint8_t mc146818write_todclk[] = {EXPECT_WORD_VAL(0x240E0086), EXPECT_WORD_VAL(0xA0AE002C), EXPECT_WORD_VAL(0x240F0070), EXPECT_WORD_VAL(0xA0AF0028), EXPECT_WORD_VAL(0xA0AA0000), EXPECT_WORD_VAL(0xA0A90008), EXPECT_WORD_VAL(0xA0A80010), EXPECT_WORD_VAL(0xA0A7001C), EXPECT_WORD_VAL(0xA0A60020), EXPECT_WORD_VAL(0x240B0048), EXPECT_WORD_VAL(0xA0AB0024), EXPECT_WORD_VAL(0x24180028), EXPECT_WORD_VAL(0xA0B80028), EXPECT_WORD_VAL(0x24190046), EXPECT_WORD_VAL(0xA0B9002C), EXPECT_WORD_VAL(0x90AC00F8), EXPECT_NOP, EXPECT_WORD_VAL(0x358D0001), EXPECT_WORD_VAL(0xA0AD00F8), EXPECT_DONE };
		
		at = ultrixPatchFind(code, nWords - 19, mc146818write_todclk);
		if (at) {
			pr("_mc146818write_todclk piece @ 0x%08x\n", at);
			
			at[11] = 0x24180020 + 34 - __builtin_clz(8192 / ADJUST_HZ_REAL_TO);
		}
		else
			return false;
	#endif
	
	return true;
}

static bool ultrixPatchData(uint32_t *data, uint32_t nWords)
{
	uint32_t *at;
	
	(void)at;
	(void)data;
	(void)nWords;
	
	#ifdef ADJUST_HZ_REPORTED_TO
		
		static const uint8_t machine_data_kn01[] = {EXPECT_WORD_VAL(0x10), 
			
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, EXPECT_ANYTHING, 
			
			EXPECT_ZERO, EXPECT_ZERO, EXPECT_WORD_VAL(0x00004e20), EXPECT_WORD_VAL(0x00000100), EXPECT_WORD_VAL(0x04000000), EXPECT_WORD_VAL(0x00000001), EXPECT_ZERO, EXPECT_DONE};
			
		at = ultrixPatchFind(data, nWords - 48, machine_data_kn01);
		if (at) {
			pr("kn01 machine data @ 0x%08x\n", at);
			at[44] = ADJUST_HZ_REPORTED_TO;
		}
		else
			return false;
		
	#endif
	
	return true;
}


bool isValidKernelName(const char *name)
{
	return !strcasecmp(name, "vmunix");
}

void loadOS(struct VfsFile *f, uint32_t fileSize, uint32_t magic, const struct DecPromVectors *vecs)
{
	static const char *argv[] = {
			"vmunix",
			"-a",		//-s for single user mode, "-a" for normal boot
	
	/*	-- this is only used if we pass in vectors, which ds2100 does not
			"0/rz6",	//first char is turbochannel slot used (we have none, so we use 0)
						//next must come a slash
						//then boot device type (for us - a disk so "rz", tape would be "tz")
						//then device scsi ID
						//to set the partition name to use (if not "a") we need to pass two more
						//	params: "-b" and then the letter of the partition
	*/
	
			0	//MUST be here!
	};
	
	struct UltrixOptionalHdr oh;
	struct CoffHdr chdr;
	uint_fast16_t i;
	
	if (sizeof(chdr) != vfsFileRead(f, &chdr, sizeof(chdr))) {
		
		pr("cannot read coff header\n");
		return;
	}
	
	if (!(chdr.flags & COFF_FILE_FLAG_EXECUTABLE) || chdr.optHdrSz != sizeof(struct UltrixOptionalHdr) ||
			sizeof(struct CoffHdr) + sizeof(struct UltrixOptionalHdr) + sizeof(struct CoffSection) * chdr.nSects > fileSize ||
			chdr.symtabOffst >= fileSize || chdr.nSymbols != sizeof(struct CoffMipsSymTabTab))
		fatal("Kernel not a valid MIPS COFF file\n");
	
	if (VERBOSE) {
		
		pr("CH:\r\n");
		pr(" %u sects\r\n", chdr.nSects);
		pr(" symtabOffst 0x%08x\r\n", chdr.symtabOffst);
		pr(" nSymbols %u\r\n", chdr.nSymbols);
		pr(" optHdrSz %u\r\n", chdr.optHdrSz);
	}
	
	if (sizeof(oh) != vfsFileRead(f, &oh, sizeof(oh))) {
		
		pr("cannot read coff 'optional'\n");
		return;
	}
	
	if (VERBOSE) {
		
		pr("OH:\r\n");
		pr(" textSz 0x%08x\r\n", oh.textSz);
		pr(" dataSz 0x%08x\r\n", oh.dataSz);
		pr(" bssSz 0x%08x\r\n", oh.bssSz);
		pr(" entrypt 0x%08x\r\n", oh.entrypt);
		pr(" textBase 0x%08x\r\n", oh.textBase);
		pr(" dataBase 0x%08x\r\n", oh.dataBase);
		pr(" bssBase 0x%08x\r\n", oh.bssBase);
		pr(" gpValue 0x%08x\r\n", oh.gpValue);
	}
	
	if (oh.magic != ULTRIX_OPT_HDR_MAGIC) {
		
		pr("corrupted optional header");
		return;
	}
	if (oh.version < ULTRIX_OPT_HDR_MIN_VER) {
		
		pr("optional header version to low");
		return;
	}

	pr("%u sections, flags: 0x%04x, entry at 0x%08x, GP 0x%08x\n", chdr.nSects, chdr.flags, oh.entrypt, oh.gpValue);

	for (i = 0; i < chdr.nSects; i++) {
		
		uint32_t pos = sizeof(struct CoffHdr) + chdr.optHdrSz + i * sizeof(struct CoffSection);
		struct CoffSection sec;
		
		if (!vfsFileSeek(f, pos) || sizeof(sec) != vfsFileRead(f, &sec, sizeof(sec))) {
			
			pr("cannot read section header %u\n", i);
			return;
		}
		
		pr("[%u]: '%8s' %08xh->%08xh+%08xh@%06xh (R: %u@%06xh L: %u@%06xh) F: %08xh\r\n", i,
			sec.name, sec.va, sec.pa, sec.size, sec.fileDataOfst,
			sec.nRelocs, sec.fileRelocOfst,
			sec.nLineNoEntries, sec.fileLineNoOfst,
			sec.flags);
		
		if (sec.va != sec.pa) {
			
			pr("addresses do not match");
			return;
		}
		
		if (sec.nRelocs) {
			
			pr("relocs not supported");
			return;
		}
		
		if (!sec.va)	//not loaded
			continue;

		if (!sec.fileDataOfst) {
			
			pr("  zeroing...\n");
			zeroWithProgress((void*)sec.va, sec.size);
		}
		else {
					
			if (sec.fileDataOfst + sec.size > fileSize) {
				
				pr("image truncted?");
				return;
			}
			
			if (!vfsFileSeek(f, sec.fileDataOfst)) {
				
				pr("cannot seek to data\n");
				return;
			}
			
			if (!copyWithProgress((void*)sec.va, f, sec.size)) {
				
				pr("cannot load data\n");
				return;
			}
			
			if (i == 0) { 		//[0] section assumed to be ".text" we could compare but that wastes valuable code space
				
				if (!ultrixPatchCode((void*)sec.va, sec.size / sizeof(uint32_t)))
					fatal("cannot patch ultrix %s\n", "code");
			}
			else if (i == 2) {	//[2] section is ".data"
				
				if (!ultrixPatchData((void*)sec.va, sec.size / sizeof(uint32_t)))
					fatal("cannot patch ultrix %s\n", "data");
			}
		}
	}

	callWithGpVal(sizeof(argv) / sizeof(*argv) - 1, argv, magic, vecs, oh.gpValue, oh.entrypt);
}

//data patern including HZ of 256: 